#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for 0D Boiler heat exchanger model.

Author: Miguel Zamarripa
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value,
                           units as pyunits)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock

from idaes.generic_models.properties import iapws95

# import ideal flue gas prop pack
from idaes.power_generation.properties import FlueGasParameterBlock
# Import Power Plant HX Unit Model
from idaes.power_generation.unit_models.boiler_heat_exchanger import (
        BoilerHeatExchanger, TubeArrangement, DeltaTMethod)

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(default={
        "side_1_property_package": m.fs.prop_steam,
        "side_2_property_package": m.fs.prop_fluegas,
        "has_pressure_change": True,
        "has_holdup": False,
        "delta_T_method": DeltaTMethod.counterCurrent,
        "tube_arrangement": TubeArrangement.inLine,
        "side_1_water_phase": "Liq",
        "has_radiation": True})

    # Check unit config arguments
    # There are 8 to 10 arguments since you can add a side 1 and 2 config by
    # side_1, side_2, or whatever the user named them
    assert len(m.fs.unit.config) >= 12 and len(m.fs.unit.config) <= 16

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.delta_T_method == \
        DeltaTMethod.counterCurrent


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_boiler_hx():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(default={
        "side_1_property_package": m.fs.prop_steam,
        "side_2_property_package": m.fs.prop_fluegas,
        "has_pressure_change": True,
        "has_holdup": False,
        "delta_T_method": DeltaTMethod.counterCurrent,
        "tube_arrangement": TubeArrangement.inLine,
        "side_1_water_phase": "Liq",
        "has_radiation": True})

    #   Set inputs
    h = value(iapws95.htpx(773.15*pyunits.K, 2.5449e7*pyunits.Pa))
    print(h)
    m.fs.unit.side_1_inlet.flow_mol[0].fix(24678.26)   # mol/s
    m.fs.unit.side_1_inlet.enth_mol[0].fix(h)           # J/mol
    m.fs.unit.side_1_inlet.pressure[0].fix(2.5449e7)    # Pascals

    # FLUE GAS Inlet from Primary Superheater
    FGrate = 28.3876e3*0.18  # mol/s equivalent of ~1930.08 klb/hr
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "H2O"].fix(FGrate*8.69/100)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "CO2"].fix(FGrate*14.49/100)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "N2"].fix(FGrate*74.34/100)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "O2"].fix(FGrate*2.47/100)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "NO"].fix(FGrate*0.0006)
    m.fs.unit.side_2_inlet.flow_mol_comp[0, "SO2"].fix(FGrate*0.002)
    m.fs.unit.side_2_inlet.temperature[0].fix(1102.335)
    m.fs.unit.side_2_inlet.pressure[0].fix(100145)

    # Primary Superheater
    ITM = 0.0254  # inch to meter conversion
    m.fs.unit.tube_di.fix((2.5-2*0.165)*ITM)
    m.fs.unit.tube_thickness.fix(0.165*ITM)
    m.fs.unit.pitch_x.fix(3*ITM)
    # gas path transverse width 54.78 ft / number of columns
    m.fs.unit.pitch_y.fix(54.78/108*12*ITM)
    m.fs.unit.tube_length.fix(53.13*12*ITM)
    m.fs.unit.tube_nrow.fix(20*2)
    m.fs.unit.tube_ncol.fix(108)
    m.fs.unit.nrow_inlet.fix(4)
    m.fs.unit.delta_elevation.fix(50)
    m.fs.unit.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.unit.tube_r_fouling = 0.003131  # (0.03131 - 0.1779 h-ft^2-F/BTU)
    if m.fs.unit.config.has_radiation is True:
        m.fs.unit.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.unit.fcorrection_htc.fix(1.5)
    # correction factor for pressure drop calc tube side
    m.fs.unit.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.unit.fcorrection_dp_shell.fix(1.0)

    assert degrees_of_freedom(m) == 0

    m.fs.unit.initialize()

    results = solver.solve(m)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok
    assert value(m.fs.unit.side_1.properties_out[0].temperature) == \
        pytest.approx(588.07, 1)
    assert value(m.fs.unit.side_2.properties_out[0].temperature) == \
        pytest.approx(573.07, 1)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.integration
def test_units():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(default={
        "side_1_property_package": m.fs.prop_steam,
        "side_2_property_package": m.fs.prop_fluegas,
        "has_pressure_change": True,
        "has_holdup": False,
        "delta_T_method": DeltaTMethod.counterCurrent,
        "tube_arrangement": TubeArrangement.inLine,
        "side_1_water_phase": "Liq",
        "has_radiation": True})

    assert_units_consistent(m)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Get project version and/or git hash
"""
# stdlib
from collections import namedtuple
import importlib
import json
import logging
from pathlib import Path
import pkg_resources
import subprocess
from types import ModuleType
from typing import Union

_log = logging.getLogger(__name__)


# Exceptions


class GetVersionError(Exception):
    pass


class ModuleImportError(GetVersionError):
    pass


class PipError(GetVersionError):
    pass


class GitHashError(GetVersionError):
    pass


# Classes and functions


#: Return from `Versioned.get_info()`
VersionInfo = namedtuple("VersionInfo", "version git_hash")


class Versioned:
    """Get version information for a module.

    Sample usage::

        from idaes.dmf.getver import Versioned
        import some_module

        vrs = Versioned(some_module)
        print(vrs.version)
        print(vrs.git_hash)

    """

    #: PIP command
    PIP = "pip"

    def __init__(self, module: Union[ModuleType, str], package: str = None):
        """Constructor.

        Args:
            module: Imported module or module namespace.
            package: Package name, overrides 'module' if present
        """
        if package:
            self._name = package
            self._mod = None
        else:
            if hasattr(module, "__package__"):
                mod = module
            else:
                try:
                    mod = importlib.import_module(module)
                except ModuleNotFoundError as err:
                    raise ModuleImportError(f"Could not import module '{module}': {err}")
            self._mod = mod
            self._name = mod.__package__
        try:
            self._version = self._get_version()
        except PipError:
            raise GetVersionError(f"Could not get version for '{self._name}' from pip command '{self.PIP}'")
        if self._version is None:
            raise GetVersionError(f"Could not get version for '{self._name}'")
        self._hash = None

    def get_info(self) -> VersionInfo:
        """Get hash and version without throwing any exceptions.
        If there was an exception, the hash may be None.

        Returns:
            The Git hash (which may be None) and the version (a string)
        """
        hash = None
        try:
            hash = self.git_hash
        except GetVersionError:
            pass
        return VersionInfo(git_hash=hash, version=self.version)

    @property
    def name(self) -> str:
        """Get the name of the module.

        Returns:
            Module name.
        """
        return self._name

    @property
    def version(self) -> str:
        """Get the version.

        Return:
            The version as a string, e.g. "2.1.0"
        """
        return self._version

    @property
    def git_hash(self) -> str:
        """Get the git hash.

        Returns:
            Git hash as a string of hexadecimal digits, e.g. "a5327b889b4606792cd31194f939c0d358e68e43"

        Raises:
            GitHashError
        """

        if self._hash is None:
            self._hash = self._get_git_hash()
        return self._hash

    # -- Protected methods -- #

    def _get_version(self):
        version = None
        try:
            version = pkg_resources.get_distribution(self._name).version
        except Exception as err:
            _log.warning(f"Could not get version from pkg_resources.get_distribution; trying pip ({self.PIP}). "
                         f"Error message was: {err}")
            status, output = subprocess.getstatusoutput(f"{self.PIP} list")
            if status != 0:
                raise PipError(f"Shell command '{self.PIP} list' failed")
            for s in output.split("\n"):
                if s.startswith(self._name):
                    version = s.split()[1]
                    break
            if version is None:
                raise PipError("Could not find version in pip output")
        return version

    def _get_git_hash(self):
        module_path = self._get_module_path()
        repo, git_hash = None, None
        # (1) try .git in module directory
        _log.debug(f"Looking for .git/ORIG_HEAD in module path '{module_path.parent}'")
        for git_head in module_path.parent.glob(r"**/.git/ORIG_HEAD"):
            _log.debug(f"Found ORIG_HEAD at {git_head}")
            repo = git_head.parent
            git_hash = git_head.open().read().strip()
            break
        if repo:
            _log.info(f"Get Git hash from repository rooted at: {repo.parent}")
        else:
            # (2) try installed dir
            _log.info(
                f"Get Git hash from installed module ({module_path}) dist-info directory"
            )
            # XXX: Find a less fragile way to do this (e.g. with setuptools)
            path = (
                module_path.parent
                / f"{self._name}-{self.version}.dist-info"
                / "direct_url.json"
            )
            if not path.exists():
                raise GitHashError(f"Cannot find magic file with Git hash: {path}")
            data = json.load(path.open())
            try:
                git_hash = data["vcs_info"]["commit_id"]
            except KeyError:
                raise GitHashError(f"Cannot find commit_id in metadata: {data}")
        _log.debug(f"Got git hash: {git_hash}")
        return git_hash

    def _get_module_path(self) -> Path:
        if self._mod is None:  # package name, use pkg_resources
            return Path(pkg_resources.get_distribution(self._name).module_path)
        paths = getattr(self._mod, "__path__", [])
        if paths:
            return Path(paths[0])
        loader = self._mod.__loader__
        if hasattr(loader, "path"):
            return Path(loader.path)
        raise ModuleImportError(f"Cannot get path for module: {self._mod}")


def get_version_info(module: Union[ModuleType, str] = None, package: str = None) -> VersionInfo:
    """Get version info for a module.

    This is simply syntactic sugar around the `Versioned` object.

    Args:
        module: The same as accepted by the :class:`Versioned` constructor.
        package: If you have a package name instead of a module, use this instead (ignore 'module' if given)

    Returns:
        The same as :meth:`Versioned.get_info()` for the given module.
    """
    if package and module:
        _log.warning("Since 'package' argument is given to get_version_info(), 'module' argument will be ignored")
    vv = Versioned(module, package=package)
    return vv.get_info()


# If run from command-line, provide hash of a module
if __name__ == "__main__":
    import argparse
    logging.basicConfig()
    prs = argparse.ArgumentParser()
    prs.add_argument("module")
    args = prs.parse_args()
    _log.setLevel(logging.DEBUG)
    try:
        info = get_version_info(args.module)
    except ModuleImportError:
        info = get_version_info(package = args.module)
    print(f"{args.module}: version={info.version} hash={info.git_hash}")


#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import sys

from idaes.surrogate.pysmo.sampling import LatinHypercubeSampling, UniformSampling, HaltonSampling, HammersleySampling, CVTSampling, SamplingMethods, FeatureScaling
import numpy as np
import pandas as pd
import pyomo.common.unittest as unittest
import pytest


class FeatureScalingTestCases(unittest.TestCase):
    """
    test_data_scaling_minmax_01: Test behaviour when input is a numpy array and with 1D array.
    test_data_scaling_minmax_02: Test behaviour when input is a numpy array and with 2D array.
    test_data_scaling_minmax_03: Test behaviour when input is a numpy array and with 3D array.
    test_data_scaling_minmax_04: Test behaviour when input is a numpy array and with 3D array with a varibale is constant.
    test_data_scaling_minmax_05: Test behaviour list input TypeError

    test_data_scaling_minmax_06: Test behaviour when input is a Pandas DataFrame and with 1D array.
    test_data_scaling_minmax_07: Test behaviour when input is a Pandas DataFrame and with 2D array.
    test_data_scaling_minmax_08: Test behaviour when input is a Pandas DataFrame and with 3D array.
    test_data_scaling_minmax_09: Test behaviour when input is a Pandas DataFrame and with 3D array with a varibale is constant.

    test_data_unscaling_minmax_01: Test behaviour when input is a numpy array and with 1D array.
    test_data_unscaling_minmax_02: Test behaviour when input is a numpy array and with 2D array.
    test_data_unscaling_minmax_03: Test behaviour when input is a numpy array and with 3D array. 
    test_data_unscaling_minmax_04: Test behaviour when input is a numpy array and with 3D array with a varibale is constant.

    test_data_unscaling_minmax_05: Test behaviour IndexError when input array size > array size
    test_data_unscaling_minmax_06: Test behaviour IndexError when input array size < array size  

    """
    def setUp(self):
        input_array_np_1d = np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
        input_array_np_2d = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        input_array_np_3d = np.array([[0,10, 11], [1,11, 15], [2,12, 21], [3,13, 29], [4,14, 39], [5,15, 51], [6,16, 65], [7,17, 81], [8,18, 99], [9,19, 119]])
        input_array_np_3d_constant = np.array([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])

        input_array_pd_1d = pd.DataFrame([[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
        input_array_pd_2d = pd.DataFrame([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        input_array_pd_3d = pd.DataFrame([[0,10, 11], [1,11, 15], [2,12, 21], [3,13, 29], [4,14, 39], [5,15, 51], [6,16, 65], [7,17, 81], [8,18, 99], [9,19, 119]])
        input_array_pd_3d_constant = pd.DataFrame([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])

        self.test_data_numpy_1d = input_array_np_1d
        self.test_data_numpy_2d = input_array_np_2d
        self.test_data_numpy_3d = input_array_np_3d
        self.test_data_numpy_3d_constant = input_array_np_3d_constant

        self.test_data_pandas_1d = input_array_pd_1d
        self.test_data_pandas_2d = input_array_pd_2d
        self.test_data_pandas_3d = input_array_pd_3d
        self.test_data_pandas_3d_constant = input_array_pd_3d_constant
        
    
    @pytest.mark.unit
    def test_data_scaling_minmax_01(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_numpy_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    
    @pytest.mark.unit
    def test_data_scaling_minmax_02(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_scaling_minmax_03(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_numpy_3d 
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_scaling_minmax_04(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_numpy_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_scaling_minmax_05(self):
        # TypeError with list
        input_array = self.test_data_numpy_2d.tolist()
        with pytest.raises(TypeError):
            FeatureScaling.data_scaling_minmax(input_array)



    @pytest.mark.unit
    def test_data_scaling_minmax_06(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_pandas_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    
    @pytest.mark.unit
    def test_data_scaling_minmax_07(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array =self.test_data_pandas_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_scaling_minmax_08(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_pandas_3d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = (input_array - expected_output_2) / (expected_output_3 - expected_output_2)
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_scaling_minmax_09(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_pandas_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)
    
    @pytest.mark.unit
    def test_data_unscaling_minmax_01(self):
        # 1D
        # Sample data generated from between 0 and 9
        input_array = self.test_data_numpy_1d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_02(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)
    
    @pytest.mark.unit
    def test_data_unscaling_minmax_03(self):
        # 3D
        # Sample data generated from the expression (x_1 + 1)^2 + x_2, x1: between 0 and 9, x2: between 10 and 19
        input_array = self.test_data_numpy_3d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)
    
    @pytest.mark.unit
    def test_data_unscaling_minmax_04(self):
        # 3D with constant
        # Sample data generated from the expression (x_1 + 1)^2 + 10, x1: between 0 and 9
        input_array = self.test_data_numpy_3d_constant
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

            

    @pytest.mark.unit
    def test_data_unscaling_minmax_05(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)

        min_array = np.array([[1]])
        max_array = np.array([[5]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)

    @pytest.mark.unit
    def test_data_unscaling_minmax_06(self):
        # 2D
        # Sample data generated from the expression (x_1 + 1)^2 between 0 and 9
        input_array = self.test_data_numpy_2d
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
                
        min_array = np.array([[1,2,3]])
        max_array = np.array([[5,6,7]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)

         
    
class SamplingMethodsTestCases(unittest.TestCase):
    def setUp(self):
        input_array_np_1d = np.array([[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
        input_array_np_2d = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        input_array_np_3d = np.array([[0,10, 11], [1,11, 15], [2,12, 21], [3,13, 29], [4,14, 39], [5,15, 51], [6,16, 65], [7,17, 81], [8,18, 99], [9,19, 119]])
        
        self.test_data_numpy_1d = input_array_np_1d
        self.test_data_numpy_2d = input_array_np_2d
        self.test_data_numpy_3d = input_array_np_3d
    """
   
    test_nearest_neighbour_01: Test behaviour with array N x d = (10,3) data and a=[-0.5,1]
    test_nearest_neighbour_02: Test behaviour with array N x d = (10,2) data and a=[-0.5]
    test_nearest_neighbour_03: Test behaviour with array N x d = (10,1) data and a=[]
    test_nearest_neighbour_04: working even with diffent input size
    test_nearest_neighbour_05: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3) 
    
    test_points_selection_01: Test behaviour with array N x d = (10,3) data and a=[[-0.5,10],[10,100]]
    test_points_selection_02: Test behaviour with array N x d = (10,2) data and a=[[-0.5],[10]]
    test_points_selection_03: Test behaviour with array N x d = (10,1) data and a=[[],[]] <---both return the first element
    test_points_selection_04: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); small
    test_points_selection_05: Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    
    test_sample_point_selection_01: selection - Test behaviour with array N x d = (10,3) data and a=[[0,0],[10,19]]
    test_sample_point_selection_02: selection - Test behaviour with array N x d = (10,2) data and a=[[0],[7]]
    test_sample_point_selection_03: selection - Test behaviour with array N x d = (10,1) data and a=[[],[]] <---both return the first element, so return only 1 []
    test_sample_point_selection_04: selection - Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    test_sample_point_selection_05: selection - Test behaviour raise ValueError if dimension of point is not matching with array N x d = (10,3); large
    
    test_points_selection_06: creation - Test behaviour with points dimension should be d with array N x d = (10,3) 
    test_points_selection_07: creation - Test behaviour with points dimension should be d with array N x d = (10,2) 
    test_points_selection_08: creation - Test behaviour with points dimension should be d with array N x d = (10,1) 
    test_points_selection_09: creation - raise IndexError if dimension of point is not matching with array N x d = (10,3); small
    test_points_selection_10: creation - raise IndexError if dimension of point is not matching with array N x d = (10,3); large
    
    test_points_selection_01: Test behaviour with n = 3
    test_points_selection_02: Test behaviour with n = 1
    test_points_selection_03: Test behaviour with n = 0
    test_points_selection_04: Test behaviour with n = -1
    test_points_selection_04: Test behaviour with n = 2.9
    
    test_base_conversion_01: Test behaviour with 5 to base 2
    test_base_conversion_02: Test behaviour with 57 to base 47
    test_base_conversion_03: Test behaviour with negative base - works, returns always 0
    test_base_conversion_04: Test behaviour raise ZeroDivisionError with 0 base
    test_base_conversion_05: Test behaviour with 1 base, infinity loop
    
    test_prime_base_to_decimal_01: Test behaviour with 0.01 in base 2 to base 10
    test_prime_base_to_decimal_02: Test behaviour with 0.01 in base 20 to base 10
    test_prime_base_to_decimal_03: working with base 1
    test_prime_base_to_decimal_04: working with base -1
    
    test_data_sequencing: in the notes, (3,2) -> expected [0, 0.5, 0.75], but returns [0.  , 0.5 , 0.25]
    """
    @pytest.mark.unit
    def test_nearest_neighbour_01(self):
        input_array = self.test_data_numpy_3d
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [-0.5,1])
        np.testing.assert_array_equal(closest_point, input_array[0,:])

    @pytest.mark.unit
    def test_nearest_neighbour_02(self):
        input_array = self.test_data_numpy_2d
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [-0.5])
        np.testing.assert_array_equal(closest_point, input_array[0,:])

    @pytest.mark.unit
    def test_nearest_neighbour_03(self):
        input_array = self.test_data_numpy_1d
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [])
        np.testing.assert_array_equal(closest_point, input_array[0,:])

    @pytest.mark.unit
    def test_nearest_neighbour_04(self):
        input_array = self.test_data_numpy_3d
        closest_point = SamplingMethods.nearest_neighbour(self, input_array, [0.5])
     
    @pytest.mark.unit
    def test_nearest_neighbour_05(self):
        input_array = self.test_data_numpy_3d
        with pytest.raises(ValueError):
            closest_point = SamplingMethods.nearest_neighbour(self, input_array, [0.5,0.9,10])

    @pytest.mark.unit
    def test_points_selection_01(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[-0.5,10],
                                             [10,100]])
        SamplingClass = SamplingMethods()
        equivalent_points = SamplingClass.points_selection(input_array, generated_sample_points)
        np.testing.assert_array_equal(equivalent_points[0], input_array[0,:])
        np.testing.assert_array_equal(equivalent_points[1], input_array[-1,:])

    @pytest.mark.unit
    def test_points_selection_02(self):
        input_array = self.test_data_numpy_2d        
        generated_sample_points =  np.array([[-0.5],
                                             [10]])
        SamplingClass = SamplingMethods()
        equivalent_points = SamplingClass.points_selection(input_array, generated_sample_points)
        np.testing.assert_array_equal(equivalent_points[0], input_array[0,:])
        np.testing.assert_array_equal(equivalent_points[1], input_array[-1,:])
    
    @pytest.mark.unit
    def test_points_selection_03(self):
        input_array = self.test_data_numpy_1d        
        generated_sample_points =  np.array([[],
                                             []])
        SamplingClass = SamplingMethods()
        equivalent_points = SamplingClass.points_selection(input_array, generated_sample_points)
        np.testing.assert_array_equal(equivalent_points[0], input_array[0,:])
        np.testing.assert_array_equal(equivalent_points[1], input_array[0,:])
    
    @pytest.mark.unit
    def test_points_selection_04(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5],
                                             [10]])
        SamplingClass = SamplingMethods()
        with pytest.raises(ValueError):
            equivalent_points = SamplingClass.points_selection(input_array, generated_sample_points)
    
    @pytest.mark.unit
    def test_points_selection_05(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5,0.7,10],
                                             [10,0.9,20]])
        SamplingClass = SamplingMethods()
        with pytest.raises(ValueError):
            equivalent_points = SamplingClass.points_selection(input_array, generated_sample_points)
    
    @pytest.mark.unit
    def test_sample_point_selection_01(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0,0],
                                             [10,19]])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'selection')
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0,:])
        np.testing.assert_array_equal(unique_sample_points[1], input_array[-1,:])

    @pytest.mark.unit
    def test_sample_point_selection_02(self):
        input_array = self.test_data_numpy_2d        
        generated_sample_points =  np.array([[0],
                                             [7]])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'selection')
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0,:])
        np.testing.assert_array_equal(unique_sample_points[1], input_array[-1,:])

    @pytest.mark.unit
    def test_sample_point_selection_03(self):
        input_array = self.test_data_numpy_1d        
        generated_sample_points =  np.array([[],
                                             []])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'selection')
        np.testing.assert_array_equal(unique_sample_points[0], input_array[0,:])
    
    @pytest.mark.unit
    def test_sample_point_selection_04(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5],
                                             [7]])
        SamplingClass = SamplingMethods()
        with pytest.raises(ValueError):
            unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'selection')
    
    @pytest.mark.unit
    def test_sample_point_selection_05(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5,1,10],
                                             [7,19,20]])
        SamplingClass = SamplingMethods()
        with pytest.raises(ValueError):
            unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'selection')

    @pytest.mark.unit
    def test_sample_point_selection_06(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5,11,3],
                                             [7,19,4]])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'creation')
        min_ , max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing,unique_sample_points)

    @pytest.mark.unit
    def test_sample_point_selection_07(self):
        input_array = self.test_data_numpy_2d        
        generated_sample_points =  np.array([[0.5,1],
                                             [7,19]])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'creation')
        min_ , max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing,unique_sample_points)

    @pytest.mark.unit
    def test_sample_point_selection_08(self):
        input_array = self.test_data_numpy_1d        
        generated_sample_points =  np.array([[0.5],
                                             [7]])
        SamplingClass = SamplingMethods()
        unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'creation')
        min_ , max_ = input_array[0, :], input_array[1, :]
        testing = min_ + generated_sample_points * (max_ - min_)
        np.testing.assert_array_equal(testing,unique_sample_points)
    
    @pytest.mark.unit
    def test_sample_point_selection_09(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[],
                                             []])
        SamplingClass = SamplingMethods()
        with pytest.raises(IndexError):
            unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'creation')
    
    @pytest.mark.unit
    def test_sample_point_selection_10(self):
        input_array = self.test_data_numpy_3d        
        generated_sample_points =  np.array([[0.5,1,10,11],
                                             [7,19,10,12]])
        SamplingClass = SamplingMethods()
        with pytest.raises(IndexError):
            unique_sample_points = SamplingClass.sample_point_selection(input_array, generated_sample_points,sampling_type = 'creation')

    @pytest.mark.unit
    def test_prime_number_generator_01(self):
        prime_list = SamplingMethods.prime_number_generator(self, 3)
        np.testing.assert_array_equal(prime_list, [2, 3, 5])

    @pytest.mark.unit
    def test_prime_number_generator_02(self):
        prime_list = SamplingMethods.prime_number_generator(self, 1)
        np.testing.assert_array_equal(prime_list, [2])

    @pytest.mark.unit
    def test_prime_number_generator_03(self):
        prime_list = SamplingMethods.prime_number_generator(self, 0)
        np.testing.assert_array_equal(prime_list, [])
    
    @pytest.mark.unit
    def test_prime_number_generator_04(self):
        prime_list = SamplingMethods.prime_number_generator(self, -1)
        np.testing.assert_array_equal(prime_list, [])

    @pytest.mark.unit
    def test_prime_number_generator_05(self):
        prime_list = SamplingMethods.prime_number_generator(self, 2.9)
        np.testing.assert_array_equal(prime_list, [2, 3, 5])
        
    @pytest.mark.unit
    def test_base_conversion_01(self):
        string_representation = SamplingMethods.base_conversion(self, 5, 2)
        assert string_representation == ['1', '0', '1']
    
    @pytest.mark.unit
    def test_base_conversion_02(self):
        string_representation = SamplingMethods.base_conversion(self, 57, 47)
        assert string_representation == ['1', '10']
    
    @pytest.mark.unit
    def test_base_conversion_03(self):
        string_representation = SamplingMethods.base_conversion(self, 10, -1)
        assert string_representation == ['0']
        
    @pytest.mark.unit
    def test_base_conversion_04(self):
        with pytest.raises(ZeroDivisionError):
            string_representation = SamplingMethods.base_conversion(self, 10, 0)

    # @pytest.mark.unit
    # def test_base_conversion_05(self):
    #     string_representation = SamplingMethods.base_conversion(self, 10, 1)

    @pytest.mark.unit
    def test_prime_base_to_decimal_01(self):
        string_representation = SamplingMethods.prime_base_to_decimal(self,['0', '0', '1'], 2)
        assert 0.25 == string_representation
    
    @pytest.mark.unit
    def test_prime_base_to_decimal_02(self):
        string_representation = SamplingMethods.prime_base_to_decimal(self,['0', '0', '1'], 20)
        assert 0.0025 == string_representation
    
    @pytest.mark.unit
    def test_prime_base_to_decimal_03(self):
        string_representation = SamplingMethods.prime_base_to_decimal(self,['0', '0', '1'], 1)
            
    @pytest.mark.unit
    def test_prime_base_to_decimal_04(self):
        string_representation = SamplingMethods.prime_base_to_decimal(self,['0', '0', '1'], -1)

    @pytest.mark.unit
    def test_data_sequencing(self):
        SamplingClass = SamplingMethods()
        sequence_decimal = SamplingClass.data_sequencing(3, 2)
        

class LatinHypercubeSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate LatinHypercubeSampling object with selection, default number of sample = 5
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate LatinHypercubeSampling object with selection, default number of sample
    test__init__selection_03: input numpy array - Test behaviour generate LatinHypercubeSampling object with selection, with a selected number of sample
    test__init__selection_04: input Pandas DataFrame - Test behaviour generate LatinHypercubeSampling object with selection, with a selected number of sample
    test__init__selection_05: input numpy array - Test behaviour raise exception with a selected number of sample = 0
    test__init__selection_06: input numpy array - Test behaviour raise exception with a selected number of sample = -1
    test__init__selection_07: input numpy array - Test behaviour raise exception with a selected number of sample > input size
    test__init__selection_08: input numpy array - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__selection_09: input list - Test behaviour raise ValueError with list input
    
    test__init__creation_01: input list - Test behaviour generate LatinHypercubeSampling object with default sampling_type, default number of sample = 5
    test__init__creation_02: input list - Test behaviour generate LatinHypercubeSampling object with sampling_type = creation, default number of sample = 5
    test__init__creation_03: input list - Test behaviour generate LatinHypercubeSampling object with creation with a selected number of sample
    test__init__creation_04: input list - Test behaviour raise exception with a selected number of sample = 0
    test__init__creation_05: input list - Test behaviour raise exception with a selected number of sample = -1
    test__init__creation_06: input list - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input numpy - Test behaviour raise ValueError with pandas input
    
    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string

    test_variable_sample_creation: Test behaviour, sampled values are in the range (min, max), number of samples = 5, 10, 1
    test_lhs_points_generation: Test behaviour, sampled values are in the range (0, 1) , number of samples = 5, 10, 1, 2 d
    test_random_shuffling: Test behaviour, random_shuffling = sampled values after soring, they are in the range (0, 1) , number of samples = 5, 10, 1, 2 d
    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1, 2 d
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1, 2 d
    
    """
    def setUp(self):
        input_array_np = np.array([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])
        input_array_pd = pd.DataFrame({'x1': [0,1,2,3,4,5,6,7,8,9],
                                       'x2': [10,10,10,10,10,10,10,10,10,10],
                                        'y': [11,14,19,26,35,46,59,74,91,110]})
        input_array_list = [[1,10,3],[2,11,4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list= input_array_list
        

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type="selection")
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])
    
    
    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type="selection")
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples, 5)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])
        

    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=6, sampling_type="selection")
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples,6)
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    
    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_pandas
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=6, sampling_type="selection")
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples,6)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(LHSClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=0, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=-1, sampling_type="selection")

    
    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=101, sampling_type="selection")
        
    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=1.1, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_list
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type="selection")

    
    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type=None)
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples,5)
    

    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type='creation')
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples,5)

    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        LHSClass = LatinHypercubeSampling( input_array, number_of_samples=100, sampling_type='creation')
        np.testing.assert_array_equal(LHSClass.data, input_array)
        np.testing.assert_array_equal(LHSClass.number_of_samples,100)

    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=0, sampling_type='creation')

    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=-1, sampling_type='creation')
    
    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=1.1, sampling_type='creation')
    
    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type='creation')

    @pytest.mark.unit
    def test__init__creation_08(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type='creation')


    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type=1)
    
    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=None, sampling_type='jp')

    @pytest.mark.unit
    def test_variable_sample_creation(self):
        input_array = self.test_data_numpy        
        for num_samples in [None,10,1]:
            LHSClass = LatinHypercubeSampling( input_array, number_of_samples=num_samples, sampling_type="selection")
            minimum, maximum = 10, 100
            out_var_samples = LHSClass.variable_sample_creation(minimum, maximum)
            assert (out_var_samples>=minimum).all() and (out_var_samples<=maximum).all()
            np.testing.assert_array_equal(LHSClass.number_of_samples, out_var_samples.shape[0])


    @pytest.mark.unit
    def test_lhs_points_generation(self):
        input_array = self.test_data_numpy
        for num_samples in [None,10,1]:
            LHSClass = LatinHypercubeSampling(input_array, number_of_samples=num_samples, sampling_type="selection")
            out_sample_points_vector = LHSClass.lhs_points_generation()
            assert (out_sample_points_vector>=0).all() and (out_sample_points_vector<=1).all()
            np.testing.assert_array_equal(LHSClass.number_of_samples, out_sample_points_vector.shape[0])
            np.testing.assert_array_equal(input_array.shape[1] - 1,out_sample_points_vector.shape[1])

        
    @pytest.mark.unit
    def test_random_shuffling(self):
        input_array = self.test_data_numpy
        for num_samples in [None,10,1]:
            LHSClass = LatinHypercubeSampling(input_array, number_of_samples=num_samples, sampling_type="selection")
            out_sample_points_vector = LHSClass.lhs_points_generation()
            vector_of_points = LHSClass.random_shuffling(out_sample_points_vector)
                    
            sidx1 = out_sample_points_vector.argsort(axis=0)
            out1 = out_sample_points_vector[sidx1, np.arange(sidx1.shape[1])]
            sidx2 = vector_of_points.argsort(axis=0)
            out2 = vector_of_points[sidx2, np.arange(sidx2.shape[1])]        

            assert (out_sample_points_vector>=0).all() and (out_sample_points_vector<=1).all()
            np.testing.assert_array_equal(out1, out2)
            np.testing.assert_array_equal(LHSClass.number_of_samples, out_sample_points_vector.shape[0])
            np.testing.assert_array_equal(input_array.shape[1] - 1,out_sample_points_vector.shape[1])

    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_numpy
            LHSClass = LatinHypercubeSampling(input_array, number_of_samples=num_samples, sampling_type="selection")
            unique_sample_points = LHSClass.sample_points()
            expected_testing = np.array([True]*unique_sample_points.shape[0], dtype=bool)
            out_testing = [unique_sample_points[i,:] in input_array for i in range(unique_sample_points.shape[0])]
            
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0),unique_sample_points)
            np.testing.assert_array_equal(expected_testing,out_testing)

    @pytest.mark.unit
    def test_sample_points_02(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_list
            LHSClass = LatinHypercubeSampling(input_array, number_of_samples=num_samples, sampling_type="creation")
            unique_sample_points = LHSClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:,i]
                assert (unique_sample_points[:,i]>=var_range[0]).all() and (unique_sample_points[:,i]<=var_range[1]).all()
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0).shape,unique_sample_points.shape)

class UniformSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate UniformSampling object with selection, default edge, list_of_samples_per_variable = [2,5]
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate UniformSampling object with selection, default edge, list_of_samples_per_variable = [2,5]
    test__init__selection_03: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is numpy, default edge
    test__init__selection_04: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is pandas, default edge
    test__init__selection_05: input numpy array - Test behaviour raise ValueError with a list_of_samples_per_variable < number of variables, default edge
    test__init__selection_06: input numpy array - Test behaviour raise ValueError with a list_of_samples_per_variable > number of variables, default edge
    test__init__selection_07: input numpy array - Test behaviour raise ValueError with a min(list_of_samples_per_variable) < 2, default edge
    test__init__selection_08: input numpy array - Test behaviour raise TypeError with a list_of_samples_per_variable is non integer, default edge
    test__init__selection_09: input numpy array - Test behaviour raise Exception with a list_of_samples_per_variable = [2,50], 2*50 > number of input data, default edge
    test__init__selection_10: input numpy array - Test behaviour generate UniformSampling object with selection, edge = True, list_of_samples_per_variable = [2,5]
    test__init__selection_11: input numpy array - Test behaviour generate UniformSampling object with selection, edge = True, list_of_samples_per_variable = [2,5]
    test__init__selection_12: input numpy array - Test behaviour raise Exception with edge = 1
    test__init__selection_13: input numpy array - Test behaviour raise Exception with edge = 'str'

    test__init__creation_01: input list - Test behaviour generate UniformSampling object with default sampling_type, default edge, list_of_samples_per_variable = [2,7,5]
    test__init__creation_02: input list - Test behaviour generate UniformSampling object with sampling_type = creation, default edge, list_of_samples_per_variable = [2,7,5]
    test__init__creation_03: input list - Test behaviour raise exception with a list_of_samples_per_variable = [1,7,5]
    test__init__creation_04: input list - Test behaviour raise exception with a list_of_samples_per_variable = [-1,7,5]
    test__init__creation_06: input list - Test behaviour raise exception with a list_of_samples_per_variable = [1.1,7,5] (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input Pandas - Test behaviour raise ValueError with Pandas input

    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string

    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = [2,5,9],[3,2,10],[4,2,28]   
    """
    def setUp(self):
        input_array_np = np.array([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])
        input_array_pd = pd.DataFrame({'x1': [0,1,2,3,4,5,6,7,8,9],
                                       'x2': [10,10,10,10,10,10,10,10,10,10],
                                        'y': [11,14,19,26,35,46,59,74,91,110]})
        input_array_list = [[1,10,3],[2,11,4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list= input_array_list

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling( input_array, [2,5], sampling_type="selection")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])

    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        UniClass = UniformSampling( input_array, [2,5], sampling_type="selection")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])
    
    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling( input_array, np.array([2,5]), sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling( input_array, pd.DataFrame([2,5]), sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling( input_array, [2], sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling( input_array, [2,5,5], sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling( input_array, [-2,5], sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(TypeError):
            UniClass = UniformSampling( input_array, [2.1,5], sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [2,50], sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_10(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling( input_array, [2,5], sampling_type="selection",edges=True)
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])
    
    @pytest.mark.unit
    def test__init__selection_11(self):
        input_array = self.test_data_numpy
        UniClass = UniformSampling( input_array, [2,5], sampling_type="selection",edges=False)
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples, 10)
        np.testing.assert_array_equal(UniClass.x_data, input_array[:, :-1])
    
    @pytest.mark.unit
    def test__init__selection_12(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [2,5], sampling_type="selection",edges=1)
    
    @pytest.mark.unit
    def test__init__selection_13(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [2,5], sampling_type="selection",edges='x')

    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        UniClass = UniformSampling( input_array, [2,7,5], sampling_type=None)
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples,2*7*5)
    
    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        UniClass = UniformSampling( input_array, [2,7,5], sampling_type="creation")
        np.testing.assert_array_equal(UniClass.data, input_array)
        np.testing.assert_array_equal(UniClass.number_of_samples,2*7*5)
    
    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [1,7,5], sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [-1,7,5], sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [1.1,7,5], sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            UniClass = UniformSampling( input_array, [2,5], sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            UniClass = UniformSampling( input_array, [2,5], sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [2,5], sampling_type=1)

    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            UniClass = UniformSampling( input_array, [2,5], sampling_type='jp')
    
    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [[2,5],[3,2],[4,2]]:
            input_array = self.test_data_numpy 
            UniClass = UniformSampling(input_array, num_samples, sampling_type="selection")
            unique_sample_points = UniClass.sample_points()
            expected_testing = np.array([True]*unique_sample_points.shape[0], dtype=bool)
            out_testing = [unique_sample_points[i,:] in input_array for i in range(unique_sample_points.shape[0])]
            
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0),unique_sample_points)
            np.testing.assert_array_equal(expected_testing,out_testing)

    @pytest.mark.unit
    def test_sample_points_02(self):
        input_array = self.test_data_list
        for num_samples in [[2,5,9],[3,2,10],[4,2,28]]:
            input_array = self.test_data_list
            UniClass = UniformSampling(input_array, num_samples, sampling_type="creation")
            unique_sample_points = UniClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:,i]
                assert (unique_sample_points[:,i]>=var_range[0]).all() and (unique_sample_points[:,i]<=var_range[1]).all()
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0).shape,unique_sample_points.shape)



class HammersleySamplingTestCases(unittest.TestCase):
    """
    __init__ = __init__ in LatinHypercubeSampling except Dimensionality problem:
        test__init__selection: Test behaviour with  dimensionality > 10

    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1
    
    """
    def setUp(self):
        input_array_np = np.array([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])
        input_array_np_large = np.array([[0, 10, 11,1,2,3,4,5,6,7,8,9,10], [1, 10, 14,1,2,3,4,5,6,7,8,9,10], [2, 10, 19,1,2,3,4,5,6,7,8,9,10], [3, 10, 26,1,2,3,4,5,6,7,8,9,10], [4, 10, 35,1,2,3,4,5,6,7,8,9,10], [5, 10, 46,1,2,3,4,5,6,7,8,9,10], [6, 10, 59,1,2,3,4,5,6,7,8,9,10], [7, 10, 74,1,2,3,4,5,6,7,8,9,10], [8, 10, 91,1,2,3,4,5,6,7,8,9,10], [9, 10, 110,1,2,3,4,5,6,7,8,9,10]])
        input_array_pd = pd.DataFrame({'x1': [0,1,2,3,4,5,6,7,8,9],
                                       'x2': [10,10,10,10,10,10,10,10,10,10],
                                        'y': [11,14,19,26,35,46,59,74,91,110]})
        input_array_list = [[1,10,3],[2,11,4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_numpy_large = input_array_np_large
        self.test_data_pandas = input_array_pd
        self.test_data_list= input_array_list
    
    @pytest.mark.unit
    def test__init__selection(self):
        input_array = self.test_data_numpy_large
        with pytest.raises(Exception):
            HammersleyClass = HammersleySampling( input_array, number_of_samples=None, sampling_type="selection")
    
    @pytest.mark.unit
    def test_sample_points_01(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_numpy 
            HammersleyClass = HammersleySampling(input_array, number_of_samples=num_samples, sampling_type="selection")
            unique_sample_points = HammersleyClass.sample_points()
            expected_testing = np.array([True]*unique_sample_points.shape[0], dtype=bool)
            out_testing = [unique_sample_points[i,:] in input_array for i in range(unique_sample_points.shape[0])]
            
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0),unique_sample_points)
            np.testing.assert_array_equal(expected_testing,out_testing)
    
    @pytest.mark.unit
    def test_sample_points_02(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_list
            HammersleyClass = HammersleySampling(input_array, number_of_samples=num_samples, sampling_type="creation")
            unique_sample_points = HammersleyClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:,i]
                assert (unique_sample_points[:,i]>=var_range[0]).all() and (unique_sample_points[:,i]<=var_range[1]).all()
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0).shape,unique_sample_points.shape)
    
    
class CVTSamplingTestCases(unittest.TestCase):
    """
    test__init__selection_01: input numpy array - Test behaviour generate object with selection, default number of sample = 5, default tolerance
    test__init__selection_02: input Pandas DataFrame - Test behaviour generate object with selection, default number of sample, default tolerance
    test__init__selection_03: input numpy array - Test behaviour generate object with selection, with a selected number of sample, default tolerance
    test__init__selection_04: input Pandas DataFrame - Test behaviour generate object with selection, with a selected number of sample, default tolerance
    test__init__selection_05: input numpy array - Test behaviour raise exception with a selected number of sample = 0, default tolerance
    test__init__selection_06: input numpy array - Test behaviour raise exception with a selected number of sample = -1, default tolerance
    test__init__selection_07: input numpy array - Test behaviour raise exception with a selected number of sample > input size, default tolerance
    test__init__selection_08: input numpy array - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer), default tolerance
    test__init__selection_09: input list - Test behaviour raise ValueError with list input
    test__init__selection_10: input numpy array - Test behaviour raise exception with tolerance > 0.1
    test__init__selection_11: input numpy array - Test behaviour raise Warning with tolerance < 0.9
    test__init__selection_12: input numpy array - Test behaviour tolerance   0.09
    test__init__selection_13: input numpy array - Test behaviour tolerance  -0.09
    
    test__init__creation_01: input list - Test behaviour generate object with default sampling_type, default number of sample = 5, default tolerance
    test__init__creation_02: input list - Test behaviour generate object with sampling_type = creation, default number of sample = 5, default tolerance
    test__init__creation_03: input list - Test behaviour generate object with creation with a selected number of sample, default tolerance
    test__init__creation_04: input list - Test behaviour raise exception with a selected number of sample = 0
    test__init__creation_05: input list - Test behaviour raise exception with a selected number of sample = -1
    test__init__creation_06: input list - Test behaviour raise exception with a selected number of sample = 1.1 (non-integer)
    test__init__creation_07: input numpy - Test behaviour raise ValueError with numpy input
    test__init__creation_08: input numpy - Test behaviour raise ValueError with pandas input
    test__init__creation_09: input list - Test behaviour raise exception with tolerance > 0.1
    test__init__creation_10: input list - Test behaviour raise Warning with tolerance < 0.9
    test__init__creation_11: input list - Test behaviour tolerance   0.09
    test__init__creation_12: input list - Test behaviour tolerance  -0.09
    
    test__init__creation_selection_01 - Test behaviour raise Exception with sampling_type = non string
    test__init__creation_selection_02 - Test behaviour raise Exception with sampling_type = incorrect string

    
    test_random_sample_selection_01 - Test random_sample_selection with size = (5,2)
    test_random_sample_selection_02 - Test random_sample_selection with size = (0,2)
    test_random_sample_selection_03 - Test random_sample_selection with size = (2,0)
    test_random_sample_selection_04 - Test behaviour raise ValueError with  size = (5,-1)
    test_random_sample_selection_05 - Test behaviour raise TypeError with size =(5,1.1)
        
    Tests: : Unit tests for eucl_distance, a function that evaluates the distance between two points (u, v).
        Four demonstration tests are done:
    test_eucl_distance_01 - The first test checks that the correct result is obtained when both inputs are single value arrays.
    test_eucl_distance_02 - The second test checks that the correct result is returned when both inputs are 2D vectors.
    test_eucl_distance_03 - The third test checks that the correct result is returned when both inputs are arrays of the same size. 
    test_eucl_distance_04 - The fourth test checks that the function is able to calculate the distance from a single point (n x 1 row vector) to a set of design points (supplied in an n x m array)

    Unit tests for create_centres, a function that generates new mass centroids for the design space based on McQueen's method.
        Four demonstration tests are done:
    test_create_centres_01 - The first test checks that the correct result for the new centres is returned when the counter is at its lowest value (1).
    test_create_centres_02 - The second test checks that the correct result for the new centres is returned when the counter is at an arbitrary value (10).
    test_create_centres_03 - The third test checks that the correct procedure is followed when one of the centres in initial_centres has no close design point to it.
    test_create_centres_04 - The fourth test checks that the the approach works as expected for problems with more than two dimensions.
    
    test_sample_points_01: Test behaviour with selection, sample points are unique, all in the input array , number of samples = 5, 10, 1
    test_sample_points_02: Test behaviour with creation, sample points are unique, all in the input range (min,max) , number of samples = 5, 10, 1

    """
    def setUp(self):
        input_array_np = np.array([[0, 10, 11], [1, 10, 14], [2, 10, 19], [3, 10, 26], [4, 10, 35], [5, 10, 46], [6, 10, 59], [7, 10, 74], [8, 10, 91], [9, 10, 110]])
        input_array_pd = pd.DataFrame({'x1': [0,1,2,3,4,5,6,7,8,9],
                                       'x2': [10,10,10,10,10,10,10,10,10,10],
                                        'y': [11,14,19,26,35,46,59,74,91,110]})
        input_array_list = [[1,10,3],[2,11,4.5]]
        self.test_data_numpy = input_array_np
        self.test_data_pandas = input_array_pd
        self.test_data_list= input_array_list
    

    @pytest.mark.unit
    def test__init__selection_01(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=None, sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps,1e-7)
    
    @pytest.mark.unit
    def test__init__selection_02(self):
        input_array = self.test_data_pandas
        CVTClass = CVTSampling( input_array, number_of_samples=None, tolerance=None, sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres, 5)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps,1e-7)
    
    @pytest.mark.unit
    def test__init__selection_03(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling( input_array, number_of_samples=6, tolerance=None,sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres,6)
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps,1e-7)
    
    @pytest.mark.unit
    def test__init__selection_04(self):
        input_array = self.test_data_pandas
        CVTClass = CVTSampling( input_array, number_of_samples=6,tolerance=None, sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres,6)
        input_array = input_array.to_numpy()
        np.testing.assert_array_equal(CVTClass.x_data, input_array[:, :-1])
        np.testing.assert_array_equal(CVTClass.eps,1e-7)
    
    @pytest.mark.unit
    def test__init__selection_05(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=0,tolerance=None, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_06(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=-1,tolerance=None, sampling_type="selection")

    
    @pytest.mark.unit
    def test__init__selection_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=101,tolerance=None, sampling_type="selection")
        
    @pytest.mark.unit
    def test__init__selection_08(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=1.1,tolerance=None, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_09(self):
        input_array = self.test_data_list
        with pytest.raises(ValueError):
            CVTClass = CVTSampling( input_array, number_of_samples=None,tolerance=None, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_10(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=0.11, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_11(self):
        input_array = self.test_data_numpy
        with pytest.warns(Warning):
            CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=1e-10, sampling_type="selection")
    
    @pytest.mark.unit
    def test__init__selection_12(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=0.09, sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.eps,0.09)

    @pytest.mark.unit
    def test__init__selection_13(self):
        input_array = self.test_data_numpy
        CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=-0.09, sampling_type="selection")
        np.testing.assert_array_equal(CVTClass.eps,-0.09)
        

    @pytest.mark.unit
    def test__init__creation_01(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling( input_array, number_of_samples=None,  tolerance=None, sampling_type=None)
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres,5)
    

    @pytest.mark.unit
    def test__init__creation_02(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling( input_array, number_of_samples=None, tolerance=None,sampling_type='creation')
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres,5)

    @pytest.mark.unit
    def test__init__creation_03(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling( input_array, number_of_samples=100,tolerance=None, sampling_type='creation')
        np.testing.assert_array_equal(CVTClass.data, input_array)
        np.testing.assert_array_equal(CVTClass.number_of_centres,100)

    @pytest.mark.unit
    def test__init__creation_04(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=0, tolerance=None,sampling_type='creation')

    @pytest.mark.unit
    def test__init__creation_05(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=-1,tolerance=None, sampling_type='creation')
    
    @pytest.mark.unit
    def test__init__creation_06(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=1.1,tolerance=None, sampling_type='creation')
    
    @pytest.mark.unit
    def test__init__creation_07(self):
        input_array = self.test_data_numpy
        with pytest.raises(ValueError):
            CVTClass = CVTSampling( input_array, number_of_samples=None, tolerance=None, sampling_type='creation')

    @pytest.mark.unit
    def test__init__creation_08(self):
        input_array = self.test_data_pandas
        with pytest.raises(ValueError):
            CVTClass = CVTSampling( input_array, number_of_samples=None, tolerance=None, sampling_type='creation')

    @pytest.mark.unit
    def test__init__creation_09(self):
        input_array = self.test_data_list
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=0.11, sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_10(self):
        input_array = self.test_data_list
        with pytest.warns(Warning):
            CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=1e-10, sampling_type="creation")
    
    @pytest.mark.unit
    def test__init__creation_11(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=0.09, sampling_type="creation")
        np.testing.assert_array_equal(CVTClass.eps,0.09)
    
    @pytest.mark.unit
    def test__init__creation_11(self):
        input_array = self.test_data_list
        CVTClass = CVTSampling( input_array,number_of_samples=None, tolerance=-0.09, sampling_type="creation")
        np.testing.assert_array_equal(CVTClass.eps,-0.09)


    @pytest.mark.unit
    def test__init__creation_selection_01(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = CVTSampling( input_array, number_of_samples=None, tolerance=None,sampling_type=1)
    
    @pytest.mark.unit
    def test__init__creation_selection_02(self):
        input_array = self.test_data_numpy
        with pytest.raises(Exception):
            CVTClass = LatinHypercubeSampling( input_array, number_of_samples=None, tolerance=None,sampling_type='jp')    
    
    @pytest.mark.unit
    def test_random_sample_selection_01(self):
        size = (5,2)
        out_random_points = CVTSampling.random_sample_selection(size[0],size[1])
        assert (out_random_points>=0).all() and (out_random_points<=1).all()
        assert out_random_points.shape == size
    

    @pytest.mark.unit
    def test_random_sample_selection_02(self):
        size = (0,2)
        out_random_points = CVTSampling.random_sample_selection(size[0],size[1])
        assert (out_random_points>=0).all() and (out_random_points<=1).all()
        assert out_random_points.shape == size
    
    @pytest.mark.unit
    def test_random_sample_selection_03(self):
        size = (2,0)
        out_random_points = CVTSampling.random_sample_selection(size[0],size[1])
        assert (out_random_points>=0).all() and (out_random_points<=1).all()
        assert out_random_points.shape == size

    @pytest.mark.unit
    def test_random_sample_selection_04(self):
        size = (5,-1)
        with pytest.raises(ValueError):
            out_random_points = CVTSampling.random_sample_selection(size[0],size[1])
    
    @pytest.mark.unit
    def test_random_sample_selection_05(self):
        size = (5,1.1)
        with pytest.raises(TypeError):
            out_random_points = CVTSampling.random_sample_selection(size[0],size[1])

    @pytest.mark.unit
    def test_eucl_distance_01(self):
        u = np.array( [ [3] ])
        v = np.array( [ [5] ])
        expected_output = 2
        output = CVTSampling.eucl_distance(u, v)
        assert expected_output == output

    @pytest.mark.unit
    def test_eucl_distance_02(self):
        u = np.array([ [1, 2] ])
        v = np.array([ [3, 4] ])
        expected_output = 8 ** 0.5
        output = CVTSampling.eucl_distance(u, v)
        assert expected_output == output

    @pytest.mark.unit
    def test_eucl_distance_03(self):
        u = np.array([ [1, 2], [3, 4] ])
        v = np.array([ [5, 6], [7, 8] ])
        expected_output = np.array([32**0.5, 32**0.5])
        output = CVTSampling.eucl_distance(u, v)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_eucl_distance_04(self):
        u = np.array([ [1, 2]])
        v = np.array([ [5, 6], [7, 8] ])
        expected_output = np.array([32**0.5, 72**0.5])
        output = CVTSampling.eucl_distance(u, v)
        np.testing.assert_array_equal(expected_output, output)\
    
    @pytest.mark.unit
    def test_create_centres_01(self):
        initial_centres = np.array([ [0, 0], [1, 1] ])
        current_random_points = np.array([[0.6, 0.6], [0.3, 0.3]])
        current_centres = np.array([1, 0])
        counter = 1
        expected_output = np.array([ [0.15, 0.15], [0.8, 0.8]])
        output = CVTSampling.create_centres(initial_centres, current_random_points, current_centres, counter)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_02(self):
        initial_centres = np.array([ [0, 0], [1, 1] ])
        current_random_points = np.array([[0.6, 0.6], [0.3, 0.3]])
        current_centres = np.array([1, 0])
        counter = 10
        expected_output = np.array([ [0.3/11, 0.3/11], [10.6/11, 10.6/11]])
        output = CVTSampling.create_centres(initial_centres, current_random_points, current_centres, counter)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_03(self):
        initial_centres = np.array([ [0, 0], [1, 1] ])
        current_random_points = np.array([[0.6, 0.6], [0.8, 0.8]])
        current_centres = np.array([1, 1])
        counter = 5
        expected_output = np.array([ [0.5/6, 0.5/6], [5.7/6, 5.7/6]])
        output = CVTSampling.create_centres(initial_centres, current_random_points, current_centres, counter)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_create_centres_04(self):
        initial_centres = np.array([ [0, 0, 0], [1, 1, 1] ])
        current_random_points = np.array([[0.1, 0.1, 0.1], [0.3, 0.3, 0.3], [0.5, 0.5, 0.5], [0.7, 0.7, 0.7], [0.9, 0.9, 0.9]])
        current_centres = np.array([0, 0, 0, 1, 1])
        counter = 4
        expected_output = np.array([ [0.3/5, 0.3/5, 0.3/5], [4.8/5, 4.8/5, 4.8/5]])
        output = CVTSampling.create_centres(initial_centres, current_random_points, current_centres, counter)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.component
    def test_sample_points_01(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_numpy
            CVTClass = CVTSampling( input_array,number_of_samples=num_samples, tolerance=None, sampling_type="selection")
            unique_sample_points = CVTClass.sample_points()
            expected_testing = np.array([True]*unique_sample_points.shape[0], dtype=bool)
            out_testing = [unique_sample_points[i,:] in input_array for i in range(unique_sample_points.shape[0])]

            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0),unique_sample_points)
            np.testing.assert_array_equal(expected_testing,out_testing)

    @pytest.mark.component
    def test_sample_points_02(self):
        for num_samples in [None,10,1]:
            input_array = self.test_data_list
            CVTClass = CVTSampling(input_array, number_of_samples=num_samples, sampling_type="creation")
            unique_sample_points = CVTClass.sample_points()
            input_array = np.array(input_array)
            for i in range(input_array.shape[1]):
                var_range = input_array[:,i]
                assert (unique_sample_points[:,i]>=var_range[0]).all() and (unique_sample_points[:,i]<=var_range[1]).all()
            np.testing.assert_array_equal(np.unique(unique_sample_points, axis=0).shape,unique_sample_points.shape)

if __name__ == '__main__':
    unittest.main()
    






#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Test for functions to save/load Pyomo model state to a dict or json
"""

import unittest
import os

from pyomo.environ import *
from idaes.core.util import to_json, from_json, StoreSpec
from idaes.util.system import mkdtemp
import shutil
import pytest

__author__ = "John Eslick"


class TestModelSerialize(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dirname = mkdtemp()
        cls.fname = os.path.join(cls.dirname, "crAzYStuff1010202030.json")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dirname)

    def tearDown(self):
        try:
            os.remove(self.fname)
        except:
            pass

    def setup_model01(self, name=None):
        if name is not None:
            model = ConcreteModel(name)
        else:
            model = ConcreteModel()
        model.b = Block([1,2,3])
        a = model.b[1].a = Var(bounds=(-100, 100), initialize=2)
        b = model.b[1].b = Var(bounds=(-100, 100), initialize=20)
        model.b[1].c = Constraint(expr=b==10*a)
        a.fix(2)
        return model

    def setup_model01b(self):
        model = ConcreteModel()
        model.b = Block(["1","2","3"])
        a = model.b["1"].a = Var(bounds=(-100, 100), initialize=2)
        b = model.b["1"].b = Var(bounds=(-100, 100), initialize=20)
        model.b["1"].c = Constraint(expr=b==10*a)
        a.fix(2)
        return model

    def setup_model02(self):
        model = ConcreteModel()
        a = model.a = Param(default=1, mutable=True)
        b = model.b = Param(default=2, mutable=True)
        c = model.c = Param(initialize=4)
        x = model.x = Var([1,2], initialize={1:1.5, 2:2.5}, bounds=(-10,10))
        model.f = Objective(expr=(x[1] - a)**2 + (x[2] - b)**2)
        model.g = Constraint(expr=x[1] + x[2] - c >= 0)
        model.dual = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        return model

    def setup_model02b(self):
        model = ConcreteModel()
        model.p = Param(["a", "b"], default={"a":1, "b":2}, mutable=True)
        a = model.p["a"]
        b = model.p["b"]
        c = model.c = Param(initialize=4)
        x = model.x = Var(["1","2"], initialize={"1":1.5, "2":2.5}, bounds=(-10,10))
        model.f = Objective(expr=(x["1"] - a)**2 + (x["2"] - b)**2)
        model.g = Constraint(expr=x["1"] + x["2"] - c >= 0)
        model.dual = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        return model

    @pytest.mark.unit
    def test01_name(self):
        """
        Simple test of load save json with differnt model names
        """
        model = self.setup_model01("m1")
        model2 = self.setup_model01("m2")

        model2.b[1].a = 0.11
        model2.b[1].b = 0.11
        to_json(model, fname=self.fname, human_read=True)
        from_json(model2, fname=self.fname)
        #make sure they are right
        assert pytest.approx(20) == value(model2.b[1].b)
        assert pytest.approx(2) == value(model2.b[1].a)

    @pytest.mark.unit
    def test01(self):
        """
        Simple test of load save json
        """
        model = self.setup_model01()
        a = model.b[1].a
        b = model.b[1].b
        to_json(model, fname=self.fname, human_read=True)
        # change variable values
        a.value = 0.11
        b.value = 0.11
        a.unfix()
        model.b[1].deactivate()
        b.setlb(2)
        b.setub(4)
        # reload values
        from_json(model, fname=self.fname)
        #make sure they are right
        assert(a.fixed)
        assert(model.b[1].active)
        assert(abs(value(b) - 20) < 1e-4)
        assert(abs(value(a) - 2) < 1e-4)
        assert(abs(b.lb - -100) < 1e-4)
        assert(abs(b.ub - 100) < 1e-4)

    @pytest.mark.unit
    def test01b(self):
        """
        Simple test of load save json
        """
        model = self.setup_model01b()
        a = model.b["1"].a
        b = model.b["1"].b
        to_json(model, fname=self.fname, human_read=True)
        # change variable values
        a.value = 0.11
        b.value = 0.11
        a.unfix()
        model.b["1"].deactivate()
        b.setlb(2)
        b.setub(4)
        # reload values
        from_json(model, fname=self.fname)
        #make sure they are right
        assert(a.fixed)
        assert(model.b["1"].active)
        assert(abs(value(b) - 20) < 1e-4)
        assert(abs(value(a) - 2) < 1e-4)
        assert(abs(b.lb - -100) < 1e-4)
        assert(abs(b.ub - 100) < 1e-4)

    @pytest.mark.unit
    def test02(self):
        """Test with suffixes"""
        model = self.setup_model02()
        x = model.x
        model.dual[model.g] = 1
        model.ipopt_zL_out[x[1]] = 0
        model.ipopt_zL_out[x[2]] = 0
        model.ipopt_zU_out[x[1]] = 0
        model.ipopt_zU_out[x[2]] = 0
        to_json(model, fname=self.fname, human_read=True)
        model.x[1].value = 10
        model.x[2].value = 10
        model.dual[model.g] = 10
        model.ipopt_zL_out[x[1]] = 10
        model.ipopt_zL_out[x[2]] = 10
        model.ipopt_zU_out[x[1]] = 10
        model.ipopt_zU_out[x[2]] = 10
        from_json(model, fname=self.fname)
        assert(abs(value(x[1]) - 1.5) < 1e-5)
        assert(abs(value(x[2]) - 2.5) < 1e-5)
        assert(abs(model.dual[model.g] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[x[1]]) < 1e-5)
        assert(abs(model.ipopt_zL_out[x[2]]) < 1e-5)
        assert(abs(model.ipopt_zU_out[x[1]]) < 1e-5)
        assert(abs(model.ipopt_zU_out[x[2]]) < 1e-5)

    @pytest.mark.unit
    def test02b(self):
        """Test with suffixes"""
        model = self.setup_model02b()
        x = model.x
        model.dual[model.g] = 1
        model.ipopt_zL_out[x["1"]] = 0
        model.ipopt_zL_out[x["2"]] = 0
        model.ipopt_zU_out[x["1"]] = 0
        model.ipopt_zU_out[x["2"]] = 0
        to_json(model, fname=self.fname, human_read=True)
        model.x["1"].value = 10
        model.x["2"].value = 10
        model.dual[model.g] = 10
        model.ipopt_zL_out[x["1"]] = 10
        model.ipopt_zL_out[x["2"]] = 10
        model.ipopt_zU_out[x["1"]] = 10
        model.ipopt_zU_out[x["2"]] = 10
        model.p["a"] = 10
        model.p["b"] = 10
        from_json(model, fname=self.fname)
        assert value(model.x["1"]) == pytest.approx(1.5)
        assert value(model.x["2"]) == pytest.approx(2.5)
        assert value(model.p["a"]) == pytest.approx(1)
        assert value(model.p["b"]) == pytest.approx(2)
        assert model.dual[model.g] == pytest.approx(1)
        assert model.ipopt_zL_out[x["1"]] == pytest.approx(0)
        assert model.ipopt_zL_out[x["2"]] == pytest.approx(0)
        assert model.ipopt_zU_out[x["1"]] == pytest.approx(0)
        assert model.ipopt_zU_out[x["2"]] == pytest.approx(0)

    @pytest.mark.unit
    def test03(self):
        """
        This tests a StoreSpec object meant for initialization.  It reloads
        the saved state of variable values and whether they are fixed or
        unfixed but it only loads values for variables that were originally
        fixed. It's use would be in doing initialization that changes which
        variables are fixed and does whatever, but when the original state is
        reloaded only the original values for fixed variables are reloaded.  You
        basically end up with the same problem, just a different initial guess.
        """
        model = self.setup_model02()
        x = model.x
        x[1].fix(1)
        wts = StoreSpec.value_isfixed(only_fixed=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        x[1].unfix()
        x[1].value = 2
        x[2].value = 10
        from_json(model, fname=self.fname, wts=wts)
        assert(x[1].fixed)
        assert(abs(value(x[1]) - 1) < 1e-5)
        assert(abs(value(x[2]) - 10) < 1e-5)

    @pytest.mark.unit
    def test04(self):
        """
        Like test03, but this StoreSpec also saves/loads active/deactivated
        component attribute and parameter values.
        """
        model = self.setup_model02()
        x = model.x
        x[1].fix(1)
        wts = StoreSpec.value_isfixed_isactive(only_fixed=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        x[1].unfix()
        x[1].value = 2
        x[2].value = 10
        model.g.deactivate()
        from_json(model, fname=self.fname, wts=wts)
        assert(x[1].fixed)
        assert(abs(value(x[1]) - 1) < 1e-5)
        assert(abs(value(x[2]) - 10) < 1e-5)
        assert(model.g.active)

    @pytest.mark.unit
    def test05(self):
        """Try just saving values"""
        model = self.setup_model02()
        model.x[1].value = 1
        wts = StoreSpec.value()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].value = 3
        model.x[2].value = 6
        from_json(model, fname=self.fname, wts=wts)
        assert(abs(value(model.x[1]) - 1) < 1e-5)
        assert(abs(model.x[1].lb + 4) < 1e-5)
        assert(abs(value(model.x[2]) - 2.5) < 1e-5)
        assert(not model.g.active)

    @pytest.mark.unit
    def test06(self):
        """Try just saving bounds"""
        model = self.setup_model02()
        model.x[1].value = 1
        wts = StoreSpec.bound()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].value = 3
        model.x[2].value = 6
        from_json(model, fname=self.fname, wts=wts)
        assert(abs(value(model.x[1]) - 3) < 1e-5)
        assert(abs(model.x[1].lb + 10) < 1e-5)
        assert(abs(value(model.x[2]) - 6) < 1e-5)
        assert(not model.g.active)

    @pytest.mark.unit
    def test07(self):
        """Try just saving just if fixed"""
        model = self.setup_model02()
        model.x[1].fix(1)
        wts = StoreSpec.isfixed()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].unfix()
        model.x[2].fix(6)
        from_json(model, fname=self.fname, wts=wts)
        assert(abs(value(model.x[1]) - 1) < 1e-5)
        assert(abs(model.x[1].lb + 4) < 1e-5)
        assert(abs(value(model.x[2]) - 6) < 1e-5)
        assert(model.x[1].fixed)
        assert(not model.x[2].fixed)
        assert(not model.g.active)

    @pytest.mark.unit
    def test08(self):
        """Try just saving suffixes"""
        model = self.setup_model02()
        model.x[1].fix(1)

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        to_json(model, fname=self.fname, wts=StoreSpec.suffix())

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].unfix()
        model.x[2].fix(6)

        from_json(model, fname=self.fname, wts=StoreSpec.suffix())
        assert(abs(value(model.x[1]) - 1) < 1e-5)
        assert(abs(value(model.x[2]) - 6) < 1e-5)
        assert(not model.x[1].fixed)
        assert(model.x[2].fixed)
        assert(not model.g.active)
        assert(abs(model.dual[model.g] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[1]] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[2]] - 1) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[1]] - 1) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[2]] - 1) < 1e-5)
        assert(abs(model.x[1].lb + 4) < 1e-5)

    @pytest.mark.unit
    def test09(self):
        """Try just saving suffixes, and suffix filter"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        to_json(model, fname=self.fname, wts=wts)

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        from_json(model, fname=self.fname, wts=wts)
        assert(abs(model.dual[model.g] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[2]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[2]] - 10) < 1e-5)

    @pytest.mark.unit
    def test10(self):
        """Try just saving suffixes, and suffix filter only on write"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        to_json(model, fname=self.fname, wts=wts)

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        from_json(model, fname=self.fname, wts=StoreSpec.suffix())
        assert(abs(model.dual[model.g] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[2]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[2]] - 10) < 1e-5)

    @pytest.mark.unit
    def test11(self):
        """Try just saving suffixes, and suffix filter only on read"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1


        to_json(model, fname=self.fname, wts=StoreSpec.suffix())

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        from_json(model, fname=self.fname, wts=wts)
        assert(abs(model.dual[model.g] - 1) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zL_out[model.x[2]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[1]] - 10) < 1e-5)
        assert(abs(model.ipopt_zU_out[model.x[2]] - 10) < 1e-5)

if __name__ == '__main__':
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
These classes handle the metadata aspects of classes representing
property packages.

Implementors of property packages need to do the following:

1. Create a new class that inherits from
   :class:`idaes.core.property_base.PhysicalParameterBlock`, which in turn
   inherits from :class:`HasPropertyClassMetadata`, in this module.

2. In that class, implement the `define_metadata()` method, inherited from
   :class:`HasPropertyClassMetadata`. This method is called
   automatically, once, when the `get_metadata()` method is first invoked.
   An empty metadata object (an instance of :class:`PropertyClassMetadata`)
   will be passed in, which the method should populate with information about
   properties and default units.

Example::

    from idaes.core.property_base import PhysicalParameterBlock

    class MyPropParams(PhysicalParameterBlock):

        @classmethod
        def define_metadata(cls, meta):
            meta.add_default_units({foo.U.TIME: 'fortnights',
                                   foo.U.MASS: 'stones'})
            meta.add_properties({'under_sea': {'units': 'leagues'},
                                'tentacle_size': {'units': 'yards'}})
            meta.add_required_properties({'under_sea': 'leagues',
                                'tentacle_size': 'yards'})

        # Also, of course, implement the non-metadata methods that
        # do the work of the class.

"""

from pyomo.core.base.units_container import _PyomoUnit

from idaes.core.util.exceptions import PropertyPackageError
import idaes.logger as idaeslog

__author__ = 'Dan Gunter <dkgunter@lbl.gov>, Andrew Lee'

_log = idaeslog.getLogger(__name__)


class HasPropertyClassMetadata(object):
    """Interface for classes that have PropertyClassMetadata.
    """
    _metadata = None

    @classmethod
    def get_metadata(cls):
        """Get property parameter metadata.

        If the metadata is not defined, this will instantiate a new
        metadata object and call `define_metadata()` to set it up.

        If the metadata is already defined, it will be simply returned.

        Returns:
            PropertyClassMetadata: The metadata
        """
        if cls._metadata is None:
            pcm = PropertyClassMetadata()
            cls.define_metadata(pcm)
            cls._metadata = pcm
        return cls._metadata

    @classmethod
    def define_metadata(cls, pcm):
        """Set all the metadata for properties and units.

        This method should be implemented by subclasses.
        In the implementation, they should set information into the
        object provided as an argument.

        Args:
            pcm (PropertyClassMetadata): Add metadata to this object.

        Returns:
            None
        """
        raise NotImplementedError()


class UnitNames(object):
    """Names for recognized units.
    """
    TM = TIME = 'time'
    LN = LENGTH = 'length'
    MA = MASS = 'mass'
    AM = AMOUNT = 'amount'
    TE = TEMPERATURE = 'temperature'
    CU = CURRENT = 'current'
    LI = LUMINOUS_INTENSITY = 'luminous intensity'


_base_units = {
    'time': None,
    'length': None,
    'mass': None,
    'amount': None,
    'temperature': None,
    'current': None,
    'luminous intensity': None}


class PropertyClassMetadata(object):
    """Container for metadata about the property class, which includes
       default units and properties.

    Example usage::

            foo = PropertyClassMetadata()
            foo.add_default_units({foo.U.TIME: 'fortnights',
                                   foo.U.MASS: 'stones'})
            foo.add_properties({'under_sea': {'units': 'leagues'},
                                'tentacle_size': {'units': 'yards'}})
            foo.add_required_properties({'under_sea': 'leagues',
                                        'tentacle_size': 'yards'})

    """
    #: Alias for class enumerating supported/known unit types
    U = UnitNames

    def __init__(self):
        self._default_units = _base_units.copy()
        self._derived_units = None
        self._properties = {}
        self._required_properties = {}

    @property
    def default_units(self):
        return self._default_units

    @property
    def derived_units(self):
        # this will return a PropertyPackageError if used on a property package
        # which has not defined units.
        if self._derived_units is None:
            self._create_derived_units()
        return self._derived_units

    @property
    def properties(self):
        return self._properties

    @property
    def required_properties(self):
        return self._required_properties

    def add_default_units(self, u):
        """Add a dict with keys for the
        quantities used in the property package (as strings) and values of
        their default units as unit objects or strings.

        The quantities used by the framework are in constants
        defined in :class:`UnitNames`, aliased here in the class
        attribute `U`.

        Args:
            u (dict): Key=property, Value=units

        Returns:
            None
        """
        self._default_units.update(u)

        # Validate values. Pyomo units are all-or-nothing, so check to see that
        # this is the case
        _units = 0
        for q, u in self._default_units.items():
            if isinstance(u, _PyomoUnit):
                _units += 1
            elif u is None and (q == "luminous intensity" or q == "current"):
                # these units are infrequently used in PSE, so allow users
                # to skip these
                continue
            elif _units > 0:
                # Mix of units and non-unit objects
                raise PropertyPackageError(
                    "default_units ({}: {}): if using Pyomo Units objects, "
                    "all units must be defined using Units objects (not "
                    "compount units)."
                    .format(q, u))

        # Take opportunity to log a deprecation warning if units are not used
        if _units == 0:
            _log.warning("DEPRECATED: IDAES is moving to using Pyomo Units "
                         "when defining default units, which are used "
                         "to automatically determine units of measurement "
                         "for quantities and convert where necessary. "
                         "Users are strongly encouraged to convert their "
                         "property packages to use Pyomo Units objects.")

    def add_properties(self, p):
        """Add properties to the metadata.

        For each property, the value should be another dict which may contain
        the following keys:

        - 'method': (required) the name of a method to construct the
                    property as a str, or None if the property will be
                    constructed by default.
        - 'units': (optional) units of measurement for the property.

        Args:
            p (dict): Key=property, Value=PropertyMetadata or equiv. dict

        Returns:
            None
        """
        for k, v in p.items():
            if not isinstance(v, PropertyMetadata):
                v = PropertyMetadata(name=k, **v)
            self._properties[k] = v

    def add_required_properties(self, p):
        """Add required properties to the metadata.

        For each property, the value should be the expected units of
        measurement for the property.

        Args:
            p (dict): Key=property, Value=units

        Returns:
            None
        """
        # Using the same PropertyMetadata class as for units, but 'method'
        # will always be none
        for k, v in p.items():
            if not isinstance(v, PropertyMetadata):
                v = PropertyMetadata(name=k, units=v)
            self._required_properties[k] = v

    def get_derived_units(self, units):
        base_units = self.default_units
        if (isinstance(base_units["mass"], str) or
                base_units["mass"] is None):
            # If one entry is not a units object, then none will be
            # Backwards compatability for pre-units property packages
            return None
        else:
            return self.derived_units[units]

    def _create_derived_units(self):
        try:
            self._derived_units = {
                "time": self.default_units["time"],
                "length": self.default_units["length"],
                "mass": self.default_units["mass"],
                "amount": self.default_units["amount"],
                "temperature": self.default_units["temperature"],
                "current": self.default_units["current"],
                "luminous intensity": self.default_units["luminous intensity"],
                "area": self.default_units["length"]**2,
                "volume": self.default_units["length"]**3,
                "flow_mass": (self.default_units["mass"] *
                              self.default_units["time"]**-1),
                "flow_mole": (self.default_units["amount"] *
                              self.default_units["time"]**-1),
                "flux_mass": (self.default_units["mass"] *
                              self.default_units["time"]**-1 *
                              self.default_units["length"]**-2),
                "flux_mole": (self.default_units["amount"] *
                              self.default_units["time"]**-1 *
                              self.default_units["length"]**-2),
                "flux_energy": (self.default_units["mass"] *
                                self.default_units["time"]**-3),
                "velocity": (self.default_units["length"] *
                             self.default_units["time"]**-1),
                "acceleration": (self.default_units["length"] *
                                 self.default_units["time"]**-2),
                "density_mass": (self.default_units["mass"] *
                                 self.default_units["length"]**-3),
                "density_mole": (self.default_units["amount"] *
                                 self.default_units["length"]**-3),
                "molecular_weight": (self.default_units["mass"] /
                                     self.default_units["amount"]),
                "energy": (self.default_units["mass"] *
                           self.default_units["length"]**2 *
                           self.default_units["time"]**-2),
                "energy_mass": (self.default_units["length"]**2 *
                                self.default_units["time"]**-2),
                "energy_mole": (self.default_units["mass"] *
                                self.default_units["length"]**2 *
                                self.default_units["time"]**-2 *
                                self.default_units["amount"]**-1),
                "entropy": (self.default_units["mass"] *
                            self.default_units["length"]**2 *
                            self.default_units["time"]**-2 *
                            self.default_units["temperature"]**-1),
                "entropy_mass": (self.default_units["length"]**2 *
                                 self.default_units["time"]**-2 *
                                 self.default_units["temperature"]**-1),
                "entropy_mole": (self.default_units["mass"] *
                                 self.default_units["length"]**2 *
                                 self.default_units["time"]**-2 *
                                 self.default_units["temperature"]**-1 *
                                 self.default_units["amount"]**-1),
                "power": (self.default_units["mass"] *
                          self.default_units["length"]**2 *
                          self.default_units["time"]**-3),
                "pressure": (self.default_units["mass"] *
                             self.default_units["length"]**-1 *
                             self.default_units["time"]**-2),
                "heat_capacity_mass": (self.default_units["length"]**2 *
                                       self.default_units["time"]**-2 *
                                       self.default_units["temperature"]**-1),
                "heat_capacity_mole": (self.default_units["mass"] *
                                       self.default_units["length"]**2 *
                                       self.default_units["time"]**-2 *
                                       self.default_units["temperature"]**-1 *
                                       self.default_units["amount"]**-1),
                "heat_transfer_coefficient":
                    (self.default_units["mass"] *
                     self.default_units["time"]**-3 *
                     self.default_units["temperature"]**-1),
                "thermal_conductivity":
                    (self.default_units["mass"] *
                     self.default_units["length"] *
                     self.default_units["time"]**-3 *
                     self.default_units["temperature"]**-1),
                "gas_constant": (self.default_units["mass"] *
                                 self.default_units["length"]**2 *
                                 self.default_units["time"]**-2 *
                                 self.default_units["temperature"]**-1 *
                                 self.default_units["amount"]**-1)}
        except TypeError:
            raise PropertyPackageError(
                "{} cannot determine derived units, as property package has "
                "not defined a set of base units.".format(str(self)))


class PropertyMetadata(dict):
    """Container for property parameter metadata.

    Instances of this class are exactly dictionaries, with the
    only difference being some guidance on the values expected in the
    dictionary from the constructor.
    """

    def __init__(self, name=None, method=None, units=None):
        if name is None:
            raise TypeError('"name" is required')
        d = {'name': name, 'method': method}
        if units is not None:
            d['units'] = units
        else:
            # Adding a default "null" unit in case it is not provided by user
            d['units'] = "-"
        super(PropertyMetadata, self).__init__(d)



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""Degeneracy Hunter is a collection of utility functions to assist in mathematical
modeling in Pyomo.
"""

__author__ = "Alexander Dowling"

import pyomo.environ as pyo
from pyomo.core.expr.visitor import identify_variables
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
import numpy as np
from scipy.sparse.linalg import svds
from scipy.sparse import issparse, find

from idaes.core.util.model_statistics import large_residuals_set, variables_near_bounds_set

from pyomo.opt import SolverStatus, TerminationCondition

import matplotlib.pyplot as plt

from operator import itemgetter


class DegeneracyHunter():

    def __init__(self, block_or_jac, solver=None):
        ''' Initialize Degeneracy Hunter Object
    
        Arguments:
            block_or_jac: Pyomo model or Jacobian
            solver: Pyomo SolverFactory
            
        Notes:
            Passing a Jacobian to Degeneracy Hunter is current untested.
        
        '''
        
        block_like = False
        try:
            block_like = issubclass(block_or_jac.ctype, pyo.Block)
        except AttributeError:
            pass
        
        if block_like:
        
            # Add Pyomo model to the object
            self.block = block_or_jac
        
            # setup pynumero interface
            self.nlp = PyomoNLP(self.block)

            # calculate Jacobian of equality constraints in COO sparse matrix format
            jac_eq = self.nlp.evaluate_jacobian_eq()

            # save the Jacobian
            self.jac_eq = jac_eq
        
            # Create a list of equality constraint names            
            self.eq_con_list = PyomoNLP.get_pyomo_equality_constraints(self.nlp)
        
            self.candidate_eqns = None
        
        elif type(block_or_jac) is np.array:
        
            raise NotImplementedError("Degeneracy Hunter only currently supports analyzing a Pyomo model")
        
            # TODO: Need to refactor, document, and test support for Jacobian
            self.jac_eq = block_or_jac
            
            self.eq_con_list = None
        
        else:
            
            raise TypeError("Check the type for 'block_or_jac'")
        
        # number of equality constraints, variables
        self.n_eq = self.jac_eq.shape[0]
        self.n_var = self.jac_eq.shape[1]
        
        
        # Define default candidate equations (enumerate)
        candidate_eqns = range(self.n_eq)
                
        # Initialize solver
        if solver is None:
            # TODO: Test performance with open solvers such as cbc
            self.solver = pyo.SolverFactory('gurobi')
            self.solver.options = {'NumericFocus':3}
            
        else:
            # TODO: Make this a custom exception following IDAES standards
            # assert type(solver) is SolverFactory, "Argument solver should be type SolverFactory" 
            self.solver = solver
            
        # Create spot to store singular values
        self.s = None
        
        # Set constants for MILPs
        self.max_nu = 1E5
        self.min_nonzero_nu = 1E-5
        
        
    def check_residuals(self, tol=1e-5, print_level=2, sort=True):
        """
        Method to return a ComponentSet of all Constraint components with a
        residual greater than a given threshold which appear in a model.
        Args:
            block : model to be studied
            tol : residual threshold for inclusion in ComponentSet
            print_level: controls to extend of output to the screen
                0: nothing printed
                1: only name of constraint printed
                2: each constraint is pretty printed
                3: pretty print each constraint, then print value for included variable
            sort: sort residuals in descending order for printing
        Returns:
            A ComponentSet including all Constraint components with a residual
            greater than tol which appear in block
        """
        
        if print_level > 0:
            residual_values = large_residuals_set(self.block, tol, True)
        else:
            return large_residuals_set(self.block, tol, False)
        

        print(" ")
        if len(residual_values) > 0:
            print("All constraints with residuals larger than",tol,":")
            if print_level == 1:
                print("Count\tName\t|residual|")
        
            if sort:
                residual_values = dict(sorted(residual_values.items(), key=itemgetter(1),reverse=True))

            for i, (c,r) in enumerate(residual_values.items()):
                if print_level == 1:
                    # Basic print statement. count, constraint, residual
                    print(i,"\t",c,"\t",r)
                else:
                    # Pretty print constraint
                    print("\ncount =",i,"\t|residual| =",r)
                    c.pprint()
    
                if print_level == 2:
                    # print values and bounds for each variable in the constraint
                    print("variable\tlower\tvalue\tupper")
                    for v in identify_variables(c.body):
                        self.print_variable_bounds(v)
        else:
            print("No constraints with residuals larger than",tol,"!")
                
                
        return residual_values.keys()

    def check_variable_bounds(self,tol=1e-5, relative=False, skip_lb = False, skip_ub = False, verbose=True):
        """
        Return a ComponentSet of all variables within a tolerance of their bounds.
        Args:
            block : model to be studied
            tol : residual threshold for inclusion in ComponentSet (default = 1e-5)
            relative : Boolean, use relative tolerance (default = False)
            skip_lb: Boolean to skip lower bound (default = False)
            skip_ub: Boolean to skip upper bound (default = False)
            verbose: Boolean to toggle on printing to screen (default = True)
        Returns:
            A ComponentSet including all Constraint components with a residual
            greater than tol which appear in block
        """
        vnbs = variables_near_bounds_set(self.block, tol, relative, skip_lb, skip_ub)
        
        if verbose:
            print(" ")
            if relative:
                s = "(relative)"
            else:
                s = "(absolute)"
            if len(vnbs) > 0:

                print("Variables within",tol,s,"of their bounds:")
                print("variable\tlower\tvalue\tupper")
                for v in vnbs:
                    self.print_variable_bounds(v)
            else:
                print("No variables within",tol,s,"of their bounds.")
            
        return vnbs
    
    def check_rank_equality_constraints(self, tol=1E-6):
        """
        Method to check the rank of the Jacobian of the equality constraints
        
        Args:
            tol: Tolerance for smallest singular value (default=1E-6)
        
        Returns:
            Number of singular values less than tolerance (-1 means error)
        
        """
        
        print("\nChecking rank of Jacobian of equality constraints...")
        
        print("Model contains",self.n_eq,"equality constraints and",
                    self.n_var,"variables.")
        
        counter = 0
        if self.n_eq > 1:
            if self.s is None:
                self.svd_analysis()

            n = len(self.s)
        
            print("Smallest singular value(s):")
            for i in range(n):
                print("%.3E" % self.s[i])
                if self.s[i] < tol:
                    counter += 1
        else:
            # TODO: Make this an exception
            print("Model needs at least 2 equality constraints to check rank.")
            counter = -1
            
        return counter

    # TODO: Refactor, this should not be a staticmethod
    @staticmethod
    def _prepare_ids_milp(jac_eq, M=1E5):
        '''
        Prepare MILP to compute the irreducible degenerate set
        
        Argument:
            jac_eq Jacobian of equality constraints [matrix]
            M: largest value for nu
            
        Returns:
            m_dh: Pyomo model to calculate irreducible degenerate sets
        '''
    
        n_eq = jac_eq.shape[0]
        n_var = jac_eq.shape[1]
    
        # Create Pyomo model for irreducible degenerate set
        m_dh = pyo.ConcreteModel()

        # Create index for constraints
        m_dh.C = pyo.Set(initialize=range(n_eq))

        m_dh.V = pyo.Set(initialize=range(n_var))

        # Add variables
        m_dh.nu = pyo.Var(m_dh.C, bounds=(-M,M), initialize=1.0)
        m_dh.y = pyo.Var(m_dh.C, domain=pyo.Binary)

        # Constraint to enforce set is degenerate
        if issparse(jac_eq):
            m_dh.J = jac_eq.tocsc()
            
            def eq_degenerate(m_dh, v):
            
                # Find the columns with non-zero entries
                C_ = find(m_dh.J[:,v])[0]
                return sum(m_dh.J[c,v] * m_dh.nu[c] for c in C_) == 0
            
        else:
            m_dh.J = jac_eq
        
            def eq_degenerate(m_dh, v):
                return sum(m_dh.J[c,v] * m_dh.nu[c] for c in m_dh.C) == 0
            
        m_dh.degenerate = pyo.Constraint(m_dh.V, rule=eq_degenerate)

        def eq_lower(m_dh, c):
            return -M * m_dh.y[c] <= m_dh.nu[c]
        m_dh.lower = pyo.Constraint(m_dh.C, rule=eq_lower)

        def eq_upper(m_dh, c):
            return  m_dh.nu[c] <= M * m_dh.y[c]
        m_dh.upper= pyo.Constraint(m_dh.C, rule=eq_upper)

        m_dh.obj = pyo.Objective(expr=sum(m_dh.y[c] for c in m_dh.C))
        
        return m_dh
    
    # TODO: Refactor, this should not be a staticmethod
    @staticmethod
    def _prepare_find_candidates_milp(jac_eq, M = 1E5, m_small = 1E-5):
        '''
        Prepare MILP to find candidate equations for consider for IDS
        
        Argument:
            jac_eq Jacobian of equality constraints [matrix]
            M: maximum value for nu
            m_small: smallest value for nu to be considered non-zero
            
        Returns:
            m_fc: Pyomo model to find candidates
        '''
    
        n_eq = jac_eq.shape[0]
        n_var = jac_eq.shape[1]
    
        # Create Pyomo model for irreducible degenerate set
        m_dh = pyo.ConcreteModel()

        # Create index for constraints
        m_dh.C = pyo.Set(initialize=range(n_eq))

        m_dh.V = pyo.Set(initialize=range(n_var))

        # Specify minimum size for nu to be considered non-zero
        m_dh.m_small = m_small

        # Add variables
        m_dh.nu = pyo.Var(m_dh.C, bounds=(-M-m_small,M+m_small),initialize=1.0)
        m_dh.y_pos = pyo.Var(m_dh.C, domain=pyo.Binary)
        m_dh.y_neg = pyo.Var(m_dh.C, domain=pyo.Binary)
        m_dh.abs_nu = pyo.Var(m_dh.C, bounds=(0, M + m_small))
        
        # Positive exclusive or negative
        def eq_pos_xor_negative(m, c):
            return m.y_pos[c] + m.y_neg[c] <= 1
        
        m_dh.pos_xor_neg = pyo.Constraint(m_dh.C)

        # Constraint to enforce set is degenerate
        if issparse(jac_eq):
            m_dh.J = jac_eq.tocsc()
            
            def eq_degenerate(m_dh, v):
                if np.sum(np.abs(m_dh.J[:,v])) > 1E-6:
                    # Find the columns with non-zero entries
                    C_ = find(m_dh.J[:,v])[0]
                    return sum(m_dh.J[c,v] * m_dh.nu[c] for c in C_) == 0
                else:
                    # This variable does not appear in any constraint
                    return pyo.Constraint.Skip
                
                
            
        else:
            m_dh.J = jac_eq
        
            def eq_degenerate(m_dh, v):
                if np.sum(np.abs(m_dh.J[:,v])) > 1E-6:
                    return sum(m_dh.J[c,v] * m_dh.nu[c] for c in m_dh.C) == 0
                else:
                    # This variable does not appear in any constraint
                    return pyo.Constraint.Skip
        
        m_dh.pprint()
        
        m_dh.degenerate = pyo.Constraint(m_dh.V, rule=eq_degenerate)

        # When y_pos = 1, nu >= m_small
        # When y_pos = 0, nu >= - m_small
        def eq_pos_lower(m_dh, c):
            return m_dh.nu[c] >= -m_small + 2*m_small * m_dh.y_pos[c]
        m_dh.pos_lower = pyo.Constraint(m_dh.C, rule=eq_pos_lower)

        # When y_pos = 1, nu <= M + m_small
        # When y_pos = 0, nu <= m_small
        def eq_pos_upper(m_dh, c):
            return m_dh.nu[c] <= M * m_dh.y_pos[c] + m_small
        m_dh.pos_upper = pyo.Constraint(m_dh.C, rule=eq_pos_upper)
        
        # When y_neg = 1, nu <= -m_small
        # When y_neg = 0, nu <= m_small
        def eq_neg_upper(m_dh, c):
            return m_dh.nu[c] <= m_small - 2*m_small * m_dh.y_neg[c]
        m_dh.neg_upper = pyo.Constraint(m_dh.C, rule=eq_neg_upper)

        # When y_neg = 1, nu >= -M - m_small
        # When y_neg = 0, nu >= - m_small
        def eq_neg_lower(m_dh, c):
            return m_dh.nu[c] >= -M * m_dh.y_neg[c] - m_small
        m_dh.neg_lower = pyo.Constraint(m_dh.C, rule=eq_neg_lower)
        
        # Absolute value
        def eq_abs_lower(m_dh,c):
            return -m_dh.abs_nu[c] <= m_dh.nu[c]
        m_dh.abs_lower = pyo.Constraint(m_dh.C, rule=eq_abs_lower)
        
        def eq_abs_upper(m_dh,c):
            return m_dh.nu[c] <= m_dh.abs_nu[c]
        m_dh.abs_upper = pyo.Constraint(m_dh.C, rule=eq_abs_upper)
        
        # At least one constraint must be in the degenerate set
        m_dh.degenerate_set_nonempty = pyo.Constraint(expr=sum(m_dh.y_pos[c] + m_dh.y_neg[c] for c in m_dh.C) >= 1)

        # Minimize the L1-norm of nu
        m_dh.obj = pyo.Objective(expr=sum(m_dh.abs_nu[c] for c in m_dh.C))
        
        return m_dh

    # TODO: Refactor, this should not be a staticmethod
    @staticmethod
    def _check_candidate_ids(ids_milp, solver, c, eq_con_list=None, tee=False):
        ''' Solve MILP to check if equation 'c' is a significant component in an irreducible
            degenerate set
            
            Arguments:
                ids_milp: Pyomo model to calculate IDS
                solver: Pyomo solver (must support MILP)
                c: index for the constraint to consider [integer]
                eq_con_list: names of equality constraints. If none, use elements of ids_milp (default=None)
                tee: Boolean, print solver output (default = False)

            Returns:
                ids: either None or dictionary containing the IDS
        '''
        
        # Fix weight on candidate equation
        ids_milp.nu[c].fix(1.0)
        
        # Solve MILP
        results = solver.solve(ids_milp, tee=tee)

        ids_milp.nu[c].unfix()

        if pyo.check_optimal_termination(results):
            # We found an irreducible degenerate set
            
            # Create empty dictionary
            ids_ = {}
            
            for i in ids_milp.C:
                # Check if constraint is included
                if ids_milp.y[i]() > 0.9:
                    # If it is, save the value of nu
                    if eq_con_list is None:
                        name = i
                    else:
                        name = eq_con_list[i]
                    ids_[name] = ids_milp.nu[i]()
            return ids_
        else:
            return None
    
    # TODO: Refactor, this should not be a staticmethod
    @staticmethod
    def _find_candidate_eqs(candidates_milp, solver, eq_con_list=None, tee=False):
        ''' Solve MILP to generate set of candidate equations
            
            Arguments:
                candidates_milp: Pyomo model to calculate IDS
                solver: Pyomo solver (must support MILP)
                eq_con_list: names of equality constraints. If none, use elements of ids_milp (default=None)
                tee: Boolean, print solver output (default = False)

            Returns:
                candidate_eqns: either None or list of indicies
                degenerate_set: either None or dictionary containing the degenerate_set
        '''
        
        results = solver.solve(candidates_milp, tee=tee)
        
        if pyo.check_optimal_termination(results):
            # We found a degenerate set
            
            # Create empty dictionary
            ds_ = {}
            
            # Create empty list
            candidate_eqns = []
            
            for i in candidates_milp.C:
                # Check if constraint is included
                if candidates_milp.abs_nu[i]() > candidates_milp.m_small * 0.99:
                    # If it is, save the value of nu
                    if eq_con_list is None:
                        name = i
                    else:
                        name = eq_con_list[i]
                    ds_[name] = candidates_milp.nu[i]()
                    candidate_eqns.append(i)
                    
            return candidate_eqns, ds_
        else:
            return None, None
        
    
    def svd_analysis(self, n_smallest_sv=10):
        '''
        Perform SVD analysis of the constraint Jacobian
        
        Args:
            n_smallest_sv: number of smallest singular values to compute
            
        Returns:
            Nothing
            
        Actions:
            Stores SVD results in object
        
        '''
        
        if self.n_eq > 1:
        
            # Determine the number of singular values to compute
            # The "-1" is needed to avoid an error with svds
            n_sv = min(n_smallest_sv, min(self.n_eq, self.n_var) - 1)
            print("Computing the",n_sv,"smallest singular value(s)")
        
            # Perform SVD
            # Recall J is a n_eq x n_var matrix
            # Thus U is a n_eq x n_eq matrix
            # And V is a n_var x n_var
            # (U or V may be smaller in economy mode)
            # Thus we really only care about U
            u, s, v = svds(self.jac_eq, k = n_sv, which='SM')
        
            # Save results
            self.u = u
            self.s = s
            self.v = v
            
        else:
            print("Warning: model must contain at least 2 equality constraints to perform svd_analysis")
    
    def find_candidate_equations(self, verbose=True, tee=False):
        '''
        Solve MILP to find a degenerate set and candidate equations
        
        Args:
            verbose: Print information to the screen (default=True)
            tee: Print solver output to screen (default=True)
        
        Returns:
            ds: either None or dictionary of candidate equations
        
        '''
        
        if verbose:
            print("*** Searching for a Single Degenerate Set ***")
            print("Building MILP model...")
        self.candidates_milp = self._prepare_find_candidates_milp(self.jac_eq,
                                                                    self.max_nu,
                                                                    self.min_nonzero_nu)
        
        if verbose:
            print("Solving MILP model...")
        ce, ds = self._find_candidate_eqs(self.candidates_milp,
                                            self.solver,
                                            self.eq_con_list,
                                            tee)
        
        if ce is not None:
            self.candidate_eqns = ce
        
        return ds
        
    
    def find_irreducible_degenerate_sets(self, verbose=True, tee=False):
        """
        Compute irreducible degenerate sets
        
        Args:
            verbose: Print information to the screen (default=True)
            tee: Print solver output to screen (default=True)
        
        Returns:
            irreducible_degenerate_sets: list of irreducible degenerate sets
        
        """
        
        # If there are no candidate equations, find them!
        if not self.candidate_eqns:
            self.find_candidate_equations()
        
        irreducible_degenerate_sets = []
        
        # Check if it is empty or None
        if self.candidate_eqns:
        
            if verbose:
                print("*** Searching for Irreducible Degenerate Sets ***")
                print("Building MILP model...")
            self.dh_milp = self._prepare_ids_milp(self.jac_eq, self.max_nu)
                
            # Loop over candidate equations
            for i, c in enumerate(self.candidate_eqns):
        
                if verbose:
                    print("Solving MILP",i+1,"of",len(self.candidate_eqns),"...")
        
                # Check if equation 'c' is a major element of an IDS
                ids_ = self._check_candidate_ids(self.dh_milp,
                                                    self.solver,
                                                    c,
                                                    self.eq_con_list,
                                                    tee)
            
                if ids_ is not None:
                    irreducible_degenerate_sets.append(ids_)

            if verbose:
                for i,s in enumerate(irreducible_degenerate_sets):
                    print("\nIrreducible Degenerate Set",i)
                    print("nu\tConstraint Name")
                    for k,v in s.items():
                        print(v,"\t",k)
        else:
            print("No candidate equations. The Jacobian is likely full rank.")
        
        
        return irreducible_degenerate_sets

    ### Helper Functions
    
    # Note: This makes sense as a static method
    @staticmethod
    def print_variable_bounds(v):
        ''' Print variable, bounds, and value
        Argument:
            v: variable
        Return:
            nothing
        '''
        print(v,"\t\t",v.lb,"\t",v.value,"\t",v.ub)


#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pytest
import sys

####
# Uncomment this to collect list of all test files
# into 'test_files.txt'
#
# def pytest_collection_modifyitems(config, items):
#     output = open("test_files.txt", "w")
#     fspaths = set()
#     for item in items:
#         fspaths.add(item.fspath)
#     for p in fspaths:
#         output.write(str(p))
#         output.write("\n")
#     output.close()
####

####


REQUIRED_MARKERS = {"unit", "component", "integration"}
ALL_PLATFORMS = {"darwin", "linux", "win32"}


@pytest.hookimpl
def pytest_runtest_setup(item):
    _validate_required_markers(
        item,
        required_markers=REQUIRED_MARKERS,
        expected_count=1
    )
    _skip_for_unsupported_platforms(
        item,
        all_platforms=ALL_PLATFORMS,
        negate_tag=(lambda tag: f'no{tag}'),
    )


def _skip_for_unsupported_platforms(item, all_platforms=None, negate_tag=None):
    """
    Only run the tests for your particular platform(s), if it is marked with those platform(s)
    e.g. to mark tests for linux:

    import pytest
    @pytest.mark.linux
    def test_something():
       print("this only runs on linux")

    In addition, you can use "no<platform>" to exclude

    @pytest.mark.nowin32
    def test_something():
       print("this will not run on windows")

    The names of the platforms should match what is returned by `sys.platform`, in particular:
       Linux = 'linux'
       Windows = 'win32'
       macOS = 'darwin'
    """

    all_platforms = set(all_platforms or [])
    if negate_tag is None:
        def negate_tag(tag):
            return f'no{tag}'
    all_negated_platforms = {negate_tag(tag) for tag in all_platforms}
    item_markers = {marker.name for marker in item.iter_markers()}
    supported_platforms = all_platforms & item_markers
    excluded_platforms = all_negated_platforms & item_markers
    plat = sys.platform
    if ((excluded_platforms and negate_tag(plat) in excluded_platforms) or
        (supported_platforms and plat not in supported_platforms)):
        pytest.skip("cannot run on platform {}".format(plat))


def _validate_required_markers(item, required_markers=None, expected_count=1):
    required_markers = set(required_markers or [])
    item_markers = {marker.name for marker in item.iter_markers()}
    required_markers_on_item = item_markers & required_markers
    required_count = len(required_markers_on_item)
    reason_to_fail = None
    if required_count < expected_count:
        reason_to_fail = 'Too few required markers'
    if required_count > expected_count:
        reason_to_fail = 'Too many required markers'
    if reason_to_fail:
        msg = (
            f'{reason_to_fail} for test function "{item.name}". '
            f'Expected: {expected_count} of {required_markers}; '
            f'found: {required_markers_on_item or required_count}'
        )
        pytest.fail(msg)


####
